import socket
import time
from maggy import util
from maggy.core import rpc
from maggy.core.reporter import Reporter
from pyspark import TaskContext


def _prepare_func(app_id, run_id, map_fun, server_addr, hb_interval):

    def _wrapper_fun(iter):
        """
        Wraps the user supplied training function in order to be passed to the
        Spark Executors.

        Args:
            iter:

        Returns:

        """

        # get task context information to determine executor identifier
        partition_id, task_attempt = util.get_partition_attempt_id()

        client = rpc.Client(server_addr, partition_id,
                            task_attempt, hb_interval)
        reporter = Reporter()

        try:
            client_addr = client.client_addr

            host_port = client_addr[0] + ":" + str(client_addr[1])

            exec_spec = {}
            exec_spec['partition_id'] = partition_id
            exec_spec['task_attempt'] = task_attempt
            exec_spec['host_port'] = host_port
            exec_spec['trial_id'] = None

            print("Registering with experiment driver")
            client.register(exec_spec)

            # blocking
            # _ = client.await_reservations()

            client.start_heartbeat(reporter)

            # blocking
            trial_id, parameters = client.get_suggestion()

            while not client.done:

                reporter.set_trial_id(trial_id)

                try:
                    print("--------------------------------")
                    print("Starting Trial: {}".format(trial_id))
                    print("Parameter Combination: {}".format(parameters))
                    retval = map_fun(**parameters, reporter=reporter)
                except util.EarlyStopException as e:
                    retval = e.metric
                    print("Early Stopped Trial.")
                finally:
                    client.finalize_metric(retval, reporter)
                    print("Finished Trial: {}".format(trial_id))
                    print("Final Metric: {}".format(retval))
                    print("--------------------------------\n")

                # blocking
                trial_id, parameters = client.get_suggestion()

        except:
            raise
        finally:
            client.stop()
            client.close()

    return _wrapper_fun



import socket
from pyspark.sql import SparkSession
from pyspark import TaskContext


def _find_spark():
    """

    Returns:
        SparkSession
    """
    return SparkSession.builder.getOrCreate()

def _get_ip_address():
    """
    Simple utility to get host IP address

    Returns:
        x
    """
    try:
	    _, _, _, _, addr = socket.getaddrinfo(socket.gethostname(),
                                              None,
                                              socket.AF_INET,
                                              socket.SOCK_STREAM)[0]
	    return addr[0]
    except:
	    return socket.gethostbyname(socket.getfqdn())

def num_executors():
    """
    Get the number of executors configured for Jupyter

    Returns:
        Number of configured executors for Jupyter
    """
    sc = _find_spark().sparkContext
    try:
        return int(sc._conf.get('spark.dynamicAllocation.maxExecutors'))
    except:
        return int(sc._conf.get('spark.executor.instances'))

def get_partition_attempt_id():
    """Returns partitionId and attemptNumber of the task context, when invoked
    on a spark executor.

    PartitionId is ID of the RDD partition that is computed by this task.

    The first task attempt will be assigned attemptNumber = 0, and subsequent
    attempts will have increasing attempt numbers.

    Returns:
        partitionId, attemptNumber -- [description]
    """
    task_context = TaskContext.get()
    return task_context.partitionId(), task_context.attemptNumber()

def print_trial_store(store):
    for _, value in store.items():
        print(value.to_json())

def _time_diff(task_start, task_end):
    """
    Args:
        :task_start:
        :tast_end:

    Returns:

    """
    time_diff = task_end - task_start

    seconds = time_diff.seconds

    if seconds < 60:
        return str(int(seconds)) + ' seconds'
    elif seconds == 60 or seconds <= 3600:
        minutes = float(seconds) / 60.0
        return str(int(minutes)) + ' minutes, ' + str((int(seconds) % 60)) + ' seconds'
    elif seconds > 3600:
        hours = float(seconds) / 3600.0
        minutes = (hours % 1) * 60
        return str(int(hours)) + ' hours, ' + str(int(minutes)) + ' minutes'
    else:
        return 'unknown time'

class EarlyStopException(Exception):

    def __init__(self, metric):
        super().__init__()

        self.metric = metric



from abc import ABC, abstractmethod

from maggy.trial import Trial

class AbstractOptimizer(ABC):

    def __init__(self, num_trials, searchspace):
        # Do stuff the user shouldn't see on initialization
        self.num_trials = num_trials
        self.searchspace = searchspace
        self.trial_buffer = []

    @abstractmethod
    def initialize(self):
        pass

    @abstractmethod
    def get_suggestion(self, trial=None):
        pass

    @abstractmethod
    def finalize_experiment(self, trials):
        pass



"""
API Module for the user to include in his training code.

"""
import threading

from maggy import util

class Reporter(object):
    """
    Thread-safe store for sending a metric from executor to driver
    """

    def __init__(self):
        self.metric = None
        self.lock = threading.RLock()
        self.stop = False
        self.trial_id = None

    # report
    def broadcast(self, metric):

        with self.lock:
            # if stop == True -> raise exception to break training function
            self.metric = metric
            if self.stop:
                raise util.EarlyStopException(metric)

    def get_metric(self):

        with self.lock:
            return self.metric

    def reset(self):
        """
        Resets the reporter to the initial state in order to start a new
        experiment.
        """
        with self.lock:
            self.metric = None
            self.stop = False
            self.trial_id = None

    def early_stop(self):

        with self.lock:
            if self.metric is not None:
                self.stop = True

    def get_trial_id(self):
        with self.lock:
            return self.trial_id

    def set_trial_id(self, trial_id):
        with self.lock:
            self.trial_id = trial_id



import random

from maggy.optimizer import AbstractOptimizer
from maggy.searchspace import Searchspace
from maggy.trial import Trial


class RandomSearch(AbstractOptimizer):

    def initialize(self):

        if (Searchspace.DOUBLE not in self.searchspace.names().values() and
                Searchspace.INTEGER not in self.searchspace.names().values()):
            raise NotImplementedError(
                "Searchspace needs at least one continuous parameter for random search.")

        for _ in range(self.num_trials):

            params = {}

            for name, value in self.searchspace.names().items():

                feasible_region = self.searchspace.get(name)

                if value == Searchspace.DOUBLE:
                    params[name] = random.uniform(feasible_region[0],
                                                  feasible_region[1])
                elif value == Searchspace.INTEGER:
                    params[name] = random.randint(feasible_region[0],
                                                  feasible_region[1])
                elif value == Searchspace.DISCRETE:
                    params[name] = random.choice(feasible_region)
                elif value == Searchspace.CATEGORICAL:
                    params[name] = random.choice(feasible_region)

            self.trial_buffer.append(Trial(params))

    def get_suggestion(self, trial=None):
        if self.trial_buffer:
            return self.trial_buffer.pop()
        else:
            return None

    def finalize_experiment(self, trials):

        first = True
        results = []
        num_trials = 0

        for trial in trials:

            if trial.status == Trial.FINALIZED:

                num_trials += 1

                metric = trial.final_metric
                param_string = trial.params

                if first:
                    max_id = trial.trial_id
                    max_hp = param_string
                    max_val = metric
                    min_id = trial.trial_id
                    min_hp = param_string
                    min_val = metric
                    first = False

                if metric > max_val:
                    max_id = trial.trial_id
                    max_val = metric
                    max_hp = param_string
                if metric < min_val:
                    min_id = trial.trial_id
                    min_val = metric
                    min_hp = param_string

                results.append(metric)

        avg = sum(results)/float(len(results))

        return {'max_id': max_id, 'max_val': max_val,
                'max_hp': max_hp, 'min_id': min_id,
                'min_val': min_val, 'min_hp': min_hp,
                'avg': avg, 'num_trials': num_trials}


"""
The experiment driver implements the functionality for scheduling trials on maggy.
"""
import queue
import threading
import datetime
import json
import maggy.util as util
from maggy.optimizer import AbstractOptimizer, RandomSearch
from maggy.core import rpc
from maggy.trial import Trial
from maggy.earlystop import AbstractEarlyStop, MedianStoppingRule
from maggy.searchspace import Searchspace


class ExperimentDriver(object):

    def __init__(self, searchspace, optimizer, direction, num_trials, name, num_executors, hb_interval, es_policy, es_interval, es_min):
        # perform type checks
        if isinstance(searchspace, Searchspace):
            self.searchspace = searchspace
        else:
            raise Exception(
                "No valid searchspace. Please use maggy Searchspace class.")

        if isinstance(optimizer, str):
            if optimizer.lower() == 'randomsearch':
                self.optimizer = RandomSearch(num_trials, self.searchspace)
            else:
                raise Exception(
                    "Unknown Optimizer. Can't initialize experiment driver.")
        elif isinstance(optimizer, AbstractOptimizer):
            print("Custom Optimizer initialized.")
            self.optimizer = optimizer
        else:
            raise Exception(
                "Unknown Optimizer. Can't initialize experiment driver.")

        if isinstance(direction, str):
            if direction.lower() not in ['min', 'max']:
                raise Exception(
                    "Unknown direction. Can't initialize experiment driver.")
        else:
            raise Exception(
                "Unknown direction. Can't initialize experiment driver.")

        if isinstance(es_policy, str):
            if es_policy.lower() == 'median':
                self.earlystop_check = MedianStoppingRule.earlystop_check
            else:
                raise Exception(
                    "Unknown Early Stopping Policy. Can't initialize experiment driver.")
        elif isinstance(es_policy, AbstractEarlyStop):
            print("Custom Early Esrly Stopping policy initialized.")
            self.earlystop_check = es_policy.earlystop_check

        self.direction = direction.lower()
        self._trial_store = {}
        self._final_store = []
        self.num_executors = num_executors
        self._message_q = queue.Queue()
        self.name = name
        self.num_trials = num_trials
        self.experiment_done = False
        self.worker_done = False
        self.hb_interval = hb_interval
        self.es_interval = es_interval
        self.es_min = es_min
        self.direction = direction.lower()
        self.server = rpc.Server(num_executors)

    def init(self):

        self.server_addr = self.server.start(self)

        self.optimizer.initialize()

        self._start_worker()

    def finalize(self, job_start, job_end):

        result = self.optimizer.finalize_experiment(self._final_store)

        self.duration = util._time_diff(job_start, job_end)

        if self.direction == 'max':
            results = '\n------ ' + str(self.optimizer.__class__.__name__) + ' results ------ direction(' + self.direction + ') \n' \
                'BEST combination ' + json.dumps(result['max_hp']) + ' -- metric ' + str(result['max_val']) + '\n' \
                'WORST combination ' + json.dumps(result['min_hp']) + ' -- metric ' + str(result['min_val']) + '\n' \
                'AVERAGE metric -- ' + str(result['avg']) + '\n' \
                'Total job time ' + self.duration + '\n'
            # TODO: write to hdfs
            print(results)
        elif self.direction == 'min':
            results = '\n------ ' + str(self.optimizer.__class__.__name__) + ' results ------ direction(' + self.direction + ') \n' \
                'BEST combination ' + json.dumps(result['min_hp']) + ' -- metric ' + str(result['min_val']) + '\n' \
                'WORST combination ' + json.dumps(result['max_hp']) + ' -- metric ' + str(result['max_val']) + '\n' \
                'AVERAGE metric -- ' + str(result['avg']) + '\n' \
                'Total job time ' + self.duration + '\n'
            # TODO: write to hdfs
            print(results)

        return result

    def get_trial(self, trial_id):
        return self._trial_store[trial_id]

    def add_trial(self, trial):
        self._trial_store[trial.trial_id] = trial

    def add_message(self, msg):
        self._message_q.put(msg)

    def _start_worker(self):

        def _target_function(self):

            time_earlystop_check = datetime.datetime.now()

            while not self.worker_done:
                trial = None
                # get a message
                try:
                    msg = self._message_q.get_nowait()
                except:
                    msg = {'type': None}

                if (datetime.datetime.now() - time_earlystop_check).total_seconds() >= self.es_interval:
                    time_earlystop_check = datetime.datetime.now()

                    # pass currently running trials to early stop component
                    if len(self._final_store) > self.es_min:
                        print("Check for early stopping.")
                        to_stop = self.earlystop_check(
                            self._trial_store, self._final_store, self.direction)
                        if len(to_stop) > 0:
                            print("Trials to stop: {}".format(to_stop))
                        for trial_id in to_stop:
                            self.get_trial(trial_id).set_early_stop()

                # depending on message do the work
                # 1. METRIC
                if msg['type'] == 'METRIC':
                    self.get_trial(msg['trial_id']).append_metric(msg['data'])

                # 2. BLACK
                elif msg['type'] == 'BLACK':
                    trial = self.get_trial(msg['trial_id'])
                    with trial.lock:
                        trial.status = Trial.SCHEDULED
                        self.server.reservations.assign_trial(
                            msg['partition_id'], msg['trial_id'])
                        # print("Scheduled Trial: " + trial.to_json())

                # 3. FINAL
                elif msg['type'] == 'FINAL':
                    # set status
                    # get trial only once
                    trial = self.get_trial(msg['trial_id'])

                    # finalize the trial object
                    with trial.lock:
                        trial.status = Trial.FINALIZED
                        trial.final_metric = msg['data']
                        trial.duration = util._time_diff(
                            trial.start, datetime.datetime.now())

                    # move trial to the finalized ones
                    self._final_store.append(trial)
                    self._trial_store.pop(trial.trial_id)

                    # TODO: make json and write to HDFS

                    # assign new trial
                    trial = self.optimizer.get_suggestion(trial)
                    if trial is None:
                        self.server.reservations.assign_trial(
                            msg['partition_id'], None)
                        self.experiment_done = True
                    else:
                        with trial.lock:
                            trial.start = datetime.datetime.now()
                            trial.status = Trial.SCHEDULED
                            self.server.reservations.assign_trial(
                                msg['partition_id'], trial.trial_id)
                            self.add_trial(trial)

                # 4. REG
                elif msg['type'] == 'REG':
                    trial = self.optimizer.get_suggestion()
                    if trial is None:
                        self.experiment_done = True
                    else:
                        with trial.lock:
                            trial.start = datetime.datetime.now()
                            trial.status = Trial.SCHEDULED
                            self.server.reservations.assign_trial(
                                msg['partition_id'], trial.trial_id)
                            self.add_trial(trial)

        t = threading.Thread(target=_target_function, args=(self,))
        t.daemon = True
        t.start()

    def stop(self):
        """Stop the Driver's worker thread and server."""
        self.worker_done = True
        self.server.stop()

    def experiment_json(self):
        pass



"""
Experiment module used for running asynchronous optimization tasks.

The programming model is that you wrap the code containing the model
training inside a wrapper function.
Inside that wrapper function provide all imports and parts that make up your
experiment, see examples below. Whenever a function to run an experiment is
invoked it is also registered in the Experiments service along with the
provided information.
"""
import socket
import datetime

from maggy import util
from maggy.core import rpc
from maggy.core import trialexecutor
from maggy.core import ExperimentDriver
from maggy.trial import Trial

app_id = None
running = False

# TODO: this should be inferred from HDFS checkpoints
run_id = 0

elastic_id = 1
experiment_json = None
driver_tensorboard_hdfs_path = None


def launch(map_fun, searchspace, optimizer, direction, num_trials, name, hb_interval=1, es_policy='median', es_interval=300, es_min=10):
    """Launches a maggy experiment for hyperparameter optimization.

    Given a search space, objective and a model training procedure `map_fun`
    (black-box function), an experiment is the whole process of finding the
    best hyperparameter combination in the search space, optimizing the
    black-box function. Currently maggy supports random search and a median
    stopping rule.

    :param map_fun: User defined experiment containing the model training.
    :type map_fun: function
    :param searchspace: A maggy Searchspace object from which samples are drawn.
    :type searchspace: Searchspace
    :param optimizer: The optimizer is the part generating new trials.
    :type optimizer: str, AbstractOptimizer
    :param direction: If set to ‘max’ the highest value returned will
        correspond to the best solution, if set to ‘min’ the opposite is true.
    :type direction: str
    :param num_trials: the number of trials to evaluate given the search space,
        each containing a different hyperparameter combination
    :type num_trials: int
    :param name: A user defined experiment identifier.
    :type name: str
    :param hb_interval: The heartbeat interval in secondss from trial executor
        to experiment driver, defaults to 1
    :type hb_interval: int, optional
    :param es_policy: The earlystopping policy, defaults to 'median'
    :type es_policy: str, optional
    :param es_interval: Frequency interval in seconds to check currently
        running trials for early stopping, defaults to 300
    :type es_interval: int, optional
    :param es_min: Minimum number of trials finalized before checking for
        early stopping, defaults to 10
    :type es_min: int, optional
    :raises RuntimeError: An experiment is currently running.
    :return: A dictionary indicating the best trial and best hyperparameter
        combination with it's performance metric
    :rtype: dict
    """
    assert num_trials > 0, "number of trials should be greater than zero"

    global running
    # move run_id to the class of the optimizer
    # global run_id
    if running:
        raise RuntimeError("An experiment is currently running.")

    try:
        global app_id
        global experiment_json
        global elastic_id
        running = True

        sc = util._find_spark().sparkContext
        app_id = str(sc.applicationId)

        num_executors = util.num_executors()

        if num_executors > num_trials:
            num_executors = num_trials

        nodeRDD = sc.parallelize(range(num_executors), num_executors)

        # Make SparkUI intuitive by grouping jobs
        sc.setJobGroup("Maggy Experiment", "{}".format(name))
        print("Started Maggy Experiment: ", "{}".format(name))

        # start experiment driver
        exp_driver = ExperimentDriver(searchspace, optimizer, direction,
            num_trials, name, num_executors, hb_interval, es_policy,
            es_interval, es_min)

        exp_driver.init()

        server_addr = exp_driver.server_addr

        # Force execution on executor, since GPU is located on executor
        job_start = datetime.datetime.now()
        nodeRDD.foreachPartition(trialexecutor._prepare_func(app_id, run_id,
            map_fun, server_addr, hb_interval))
        job_end = datetime.datetime.now()

        result = exp_driver.finalize(job_start, job_end)

        print("Finished Experiment \n")

    except:
        raise
    finally:
        # cleanup spark jobs
        exp_driver.stop()
        running = False
        sc.setJobGroup("", "")

    return result




import threading
import struct
import pickle
import time
import select
import socket

from maggy import util
from maggy.trial import Trial

MAX_RETRIES = 3
BUFSIZE = 1024 * 2


class Reservations(object):
    """Thread-safe store for worker reservations.

    Needs to be thread-safe mainly because the server listener thread can add
    reservations while the experiment driver might modify something on a
    reservation.
    """

    def __init__(self, required):
        """

        Args:
            required:
        """
        self.required = required
        self.lock = threading.RLock()
        self.reservations = {}
        self.check_done = False

    def add(self, meta):
        """
        Add a reservation.

        Args:
            :meta: a dictonary of metadata about a node
        """
        with self.lock:
            self.reservations[meta['partition_id']] = {
                'host_port': meta["host_port"],
                'task_attempt': meta['task_attempt'],
                'trial_id': meta['trial_id']}

            if self.remaining() == 0:
                self.check_done = True

    def done(self):
        """Returns True if the ``required`` number of reservations have been fulfilled."""
        with self.lock:
            return self.check_done

    def get(self):
        """Get the current reservations."""
        with self.lock:
            return self.reservations

    def remaining(self):
        """Get a count of remaining/unfulfilled reservations."""
        with self.lock:
            num_registered = len(self.reservations)
            return self.required - num_registered

    def get_assigned_trial(self, partition_id):
        """Get the ``trial_id`` of the trial assigned to ``partition_id``.

        Returns None if executor with ``partition_id`` is not registered or if
        ``partition_id`` is not assigned a trial yet.

        Args:
            :partition_id: An id to identify the spark executor.

        Returns:
            trial_id
        """
        with self.lock:
            reservation = self.reservations.get(partition_id, None)
            if reservation is not None:
                return reservation.get('trial_id', None)

    def assign_trial(self, partition_id, trial_id):
        """Assigns trial with ``trial_id`` to the reservation with ``partition_id``.

        Args:
            partition_id --
            trial {[type]} -- [description]
        """
        with self.lock:
            self.reservations.get(partition_id, None)['trial_id'] = trial_id


class MessageSocket(object):
    """Abstract class w/ length-prefixed socket send/receive functions."""

    def receive(self, sock):
        """
        Receive a message on ``sock``

        Args:
            sock:

        Returns:

        """
        msg = None
        data = b''
        recv_done = False
        recv_len = -1
        while not recv_done:
            buf = sock.recv(BUFSIZE)
            if buf is None or len(buf) == 0:
                raise Exception("socket closed")
            if recv_len == -1:
                recv_len = struct.unpack('>I', buf[:4])[0]
                data += buf[4:]
                recv_len -= len(data)
            else:
                data += buf
                recv_len -= len(buf)
            recv_done = (recv_len == 0)

        msg = pickle.loads(data)
        return msg

    def send(self, sock, msg):
        """
        Send ``msg`` to destination ``sock``.

        Args:
            sock:
            msg:

        Returns:

        """
        data = pickle.dumps(msg)
        buf = struct.pack('>I', len(data)) + data
        sock.sendall(buf)


class Server(MessageSocket):
    """Simple socket server with length prefixed pickle messages"""
    reservations = None
    done = False

    def __init__(self, count):
        """

        Args:
            count:
        """
        assert count > 0
        self.reservations = Reservations(count)

    def await_reservations(self, sc, status={}, timeout=600):
        """
        Block until all reservations are received.

        Args:
            sc:
            status:
            timeout:

        Returns:

        """
        timespent = 0
        while not self.reservations.done():
            print("waiting for {} reservations."
                  .format(self.reservations.remaining()))
            # check status flags for any errors
            if 'error' in status:
                sc.cancelAllJobs()
            time.sleep(1)
            timespent += 1
            if (timespent > timeout):
                raise Exception(
                    "Timed out waiting for reservations to complete")
        print("All reservations completed")
        return self.reservations.get()

    def _handle_message(self, sock, msg, exp_driver):
        """
        Handles a  message dictionary. Expects a 'type' and 'data' attribute in
        the message dictionary.

        Args:
            sock:
            msg:

        Returns:

        """
        msg_type = msg['type']

        if 'trial_id' in msg:
            # throw away idle heartbeat messages without trial
            if msg['trial_id'] is None:
                send = {}
                send['type'] = 'OK'
                MessageSocket.send(self, sock, send)
                return
            # it can happen that executor has trial but no metric was reported
            # yet so trial_id is not None but data is None
            elif msg['trial_id'] is not None:
                if msg.get('data', None) is None:
                    send = {}
                    send['type'] = 'OK'
                    MessageSocket.send(self, sock, send)
                    return

        if msg_type == 'REG':
            # check if executor was registered before and retrieve lost trial
            lost_trial = self.reservations.get_assigned_trial(msg['partition_id'])
            if lost_trial is not None:
                # the trial or executor must have failed
                exp_driver.get_trial(lost_trial).status = Trial.ERROR
                # add a blacklist message to the worker queue
                fail_msg = {'partition_id': msg['partition_id'],
                    'type': 'BLACK',
                    'trial_id': lost_trial}
                self.reservations.add(msg['data'])
                exp_driver.add_message(fail_msg)
            else:
                # else add regular registration msg to queue
                self.reservations.add(msg['data'])
                exp_driver.add_message(msg)

            send = {}
            send['type'] = 'OK'

            MessageSocket.send(self, sock, send)
        elif msg_type == 'QUERY':

            send = {}
            send['type'] = 'QUERY'
            send['data'] = self.reservations.done()

            MessageSocket.send(self, sock, send)
        elif msg_type == 'METRIC':
            # Prepare message
            send = {}

            # lookup executor reservation to find assigned trial
            trialId = msg['trial_id']
            # get early stopping flag
            flag = exp_driver.get_trial(trialId).get_early_stop()
            # add metric msg to the exp driver queue
            exp_driver.add_message(msg)

            if flag:
                send['type'] = 'STOP'
            else:
                send['type'] = 'OK'

            MessageSocket.send(self, sock, send)
        elif msg_type == 'FINAL':
            # reset the reservation to avoid sending the same trial again
            self.reservations.assign_trial(msg['partition_id'], None)

            send = {}
            send['type'] = 'OK'

            # add metric msg to the exp driver queue
            exp_driver.add_message(msg)

            MessageSocket.send(self, sock, send)
        elif msg_type == 'GET':
            # lookup reservation to find assigned trial
            trial_id = self.reservations.get_assigned_trial(msg['partition_id'])

            # trial_id needs to be none because experiment_done can be true but
            # the assigned trial might not be finalized yet
            if exp_driver.experiment_done and trial_id is None:
                send = {}
                send['type'] = "GSTOP"
            else:
                send = {}
                send['type'] = "TRIAL"

            send['trial_id'] = trial_id

            # retrieve trial information
            if trial_id is not None:
                send['data'] = exp_driver.get_trial(trial_id).params
                exp_driver.get_trial(trial_id).status = Trial.RUNNING
            else:
                send['data'] = None

            MessageSocket.send(self, sock, send)
        else:
            # Prepare message
            send = {}
            send['type'] = "ERR"
            MessageSocket.send(self, sock, send)

    def get_assigned_trial_id(self, partition_id):
        """Returns the id of the assigned trial, given a ``partition_id``.

        Arguments:
            partition_id {[type]} -- [description]

        Returns:
            trial_id
        """
        return self.reservations.get_assigned_trial(partition_id)

    def start(self, exp_driver):
        """
        Start listener in a background thread.

        Returns:
            address of the Server as a tuple of (host, port)
        """
        server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_sock.bind(('', 0))
        server_sock.listen(10)

        # hostname may not be resolvable but IP address probably will be
        host = util._get_ip_address()
        port = server_sock.getsockname()[1]
        addr = (host, port)

        def _listen(self, sock, driver):
            CONNECTIONS = []
            CONNECTIONS.append(sock)

            while not self.done:
                read_socks, _, _ = select.select(
                    CONNECTIONS, [], [], 60)
                for sock in read_socks:
                    if sock == server_sock:
                        client_sock, client_addr = sock.accept()
                        CONNECTIONS.append(client_sock)
                        _ = client_addr
                    else:
                        try:
                            msg = self.receive(sock)
                            self._handle_message(sock, msg, driver)
                        except Exception as e:
                            _ = e
                            sock.close()
                            CONNECTIONS.remove(sock)

            server_sock.close()

        t = threading.Thread(target=_listen, args=(self, server_sock, exp_driver))
        t.daemon = True
        t.start()

        return addr

    def stop(self):
        """
        Stop the server's socket listener.
        """
        self.done = True


class Client(MessageSocket):
    """Client to register and await node reservations.

    Args:
        :server_addr: a tuple of (host, port) pointing to the Server.
    """
    def __init__(self, server_addr, partition_id, task_attempt, hb_interval):
        # socket for main thread
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect(server_addr)
        # socket for heartbeat thread
        self.hb_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.hb_sock.connect(server_addr)
        self.server_addr = server_addr
        self.done = False
        self.client_addr = (util._get_ip_address(), self.sock.getsockname()[1])
        self.partition_id = partition_id
        self.task_attempt = task_attempt
        self.hb_interval = hb_interval

    def _request(self, req_sock, msg_type, msg_data=None, trial_id=None):
        """Helper function to wrap msg w/ msg_type."""
        msg = {}
        msg['partition_id'] = self.partition_id
        msg['type'] = msg_type

        if msg_type == 'FINAL' or msg_type == 'METRIC':
            msg['trial_id'] = trial_id

        if msg_data or ((msg_data == True) or (msg_data == False)):
            msg['data'] = msg_data

        done = False
        tries = 0
        while not done and tries < MAX_RETRIES:
            try:
                MessageSocket.send(self, req_sock, msg)
                done = True
            except socket.error as e:
                tries += 1
                if tries >= MAX_RETRIES:
                    raise
                print("Socket error: {}".format(e))
                req_sock.close()
                req_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                req_sock.connect(self.server_addr)

        resp = MessageSocket.receive(self, req_sock)

        return resp

    def close(self):
        """Close the client's sockets."""
        self.sock.close()
        self.hb_sock.close()

    def register(self, registration):
        """
        Register ``registration`` with server.

        Args:
            registration:

        Returns:

        """
        resp = self._request(self.sock, 'REG', registration)
        return resp

    def await_reservations(self):
        done = False
        while not done:
            done = self._request(self.sock, 'QUERY').get('data', False)
            time.sleep(1)
        print("All executors registered: {}".format(done))
        return done

    def start_heartbeat(self, reporter):

        def _heartbeat(self, report):

            while not self.done:

                metric = report.get_metric()

                resp = self._request(self.hb_sock,
                                    'METRIC',
                                     metric,
                                     report.get_trial_id())
                _ = self._handle_message(resp, report)

                # sleep one second
                time.sleep(self.hb_interval)

        t = threading.Thread(target=_heartbeat, args=(self, reporter))
        t.daemon = True
        t.start()

        print("Started metric heartbeat")

    def get_suggestion(self):
        """Blocking call to get new parameter combination."""
        while not self.done:
            resp = self._request(self.sock, 'GET')
            trial_id, parameters = self._handle_message(resp) or (None, None)

            if trial_id is not None:
                break
            time.sleep(1)
        return trial_id, parameters

    def stop(self):
        """Stop the Clients's heartbeat thread."""
        self.done = True

    def _handle_message(self, msg, reporter=None):
        """
        Handles a  message dictionary. Expects a 'type' and 'data' attribute in
        the message dictionary.

        Args:
            sock:
            msg:

        Returns:

        """
        msg_type = msg['type']
        # if response is STOP command, early stop the training
        if msg_type == 'STOP':
            reporter.early_stop()
        elif msg_type == 'GSTOP':
            print("Stopping experiment")
            self.done = True
        elif msg_type == 'TRIAL':
            return msg['trial_id'], msg['data']
        elif msg_type == 'ERR':
            print("Stopping experiment")
            self.done = True

    def finalize_metric(self, metric, reporter):
        # make sure heartbeat thread can't send between sending final metric
        # and resetting the reporter
        with reporter.lock:
            resp = self._request(self.sock, 'FINAL', metric, reporter.get_trial_id())
            reporter.reset()
        return resp



import os
from setuptools import setup, find_packages
from maggy.version import __version__


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='maggy',
    version=__version__,
    author='Moritz Meister',
    author_email='meister.mo@gmail.com',
    description='',
    license='GNU Affero General Public License v3',
    keywords='Hyperparameter, Optimization, Auto-ML, Hops, Hadoop, TensorFlow, Spark',
    url='https://github.com/logicalclocks/maggy',
    # download_url = 'http://snurran.sics.se/hops/hops-util-py/hops-' + __version__ + '.tar.gz',
    packages=find_packages(),
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python :: 3',
    ]
)
